#define LINK_EXPR_C 1

#include <zasm/errors.h>
#include <zasm/alloc.h>
#include <zasm/link_expr.h>
#include <zasm/link_namespace.h>
#include <string.h>

ALLOC_DEFINE(link_expr, ALLOC_DEF_COUNT);

struct link_expr_t *link_expr_new(struct link_level_t *level)
{
  struct link_expr_t *val;
  
  if (!level)
    ERROR("internal error\n");

  val = link_expr_alloc_get(&(level->expr_alloc));
  link_expr_fill(val);
  return val;
}

void link_expr_fill(struct link_expr_t *expr)
{}

struct link_expr_t *link_expr_copy(struct link_expr_t *expr, struct link_level_t *to)
{
  struct link_expr_t *val;

  if (!expr)
    return NULL;

  val = link_expr_new(to);
  val->type = expr->type;
  switch(val->type)
    {
    case EXPR_OPERATOR:
      val->value.function.calc = expr->value.function.calc;
      val->value.function.arg1 = link_expr_copy(expr->value.function.arg1, to);
      val->value.function.arg2 = link_expr_copy(expr->value.function.arg2, to);
      val->value.function.arg3 = link_expr_copy(expr->value.function.arg3, to);
      break;
    case EXPR_NUMBER:
      val->value.value = expr->value.value;
      break;
    case EXPR_PAGE:
    case EXPR_VALUE:
      val->value.entry = link_level_lookup_new(to, expr->value.entry->entry.name);
      //link_entry_copy(expr->value.entry, to);
      break;
    default:
      ERROR("internal error\n");
    }
  return val;
}

struct link_expr_t *link_expr_from_expr(struct link_level_t *level, struct expr_t *expr, int addr, int page, int instr_start)
{
  struct link_expr_t *val;
  char *name;
  char *part;

  if (!expr)
    return NULL;
  val = link_expr_new(level);
  switch (expr->type)
    {
    case EXPR_OPERATOR:
      val->type = EXPR_OPERATOR;
      val->value.function.arg1 = link_expr_from_expr(level, expr->value.function.arg1, addr, page, instr_start);
      val->value.function.arg2 = link_expr_from_expr(level, expr->value.function.arg2, addr, page, instr_start);
      val->value.function.arg3 = link_expr_from_expr(level, expr->value.function.arg3, addr, page, instr_start);
      val->value.function.calc = expr->value.function.calc;
      break;
    case EXPR_NUMBER:
      val->type = EXPR_NUMBER;
      val->value.value = expr->value.value;
      break;
    case EXPR_NAME:
      name = expr->value.name.name;
      part = expr->value.name.part;
      if ((name[0] == '$') && (name[1] == 0))
	{
	  val->type = EXPR_NUMBER;
	  if (part && (0==strcmp(part, "page")))
	    val->value.value = page;
	  else if (part && (0==strcmp(part, "end")))
	    val->value.value = addr;
	  else
	    val->value.value = instr_start;
	}
      else if ((name[0] == '@') && (name[1] == 0))
	{
	  val->type = EXPR_NUMBER;
	  if (part && (0==strcmp(part, "address")))
	    val->value.value = addr;
	  else
	    val->value.value = page;
	}
      else if (part && (0==strcmp(part, "page")))
	{
	  val->type = EXPR_PAGE;
	  val->value.entry = link_level_lookup_new(level, name);
	}
      else
	{
	  val->type = EXPR_VALUE;
	  val->value.entry = link_level_lookup_new(level, name);
	}
      break;
    default:
      ERROR("internal error\n");
    }
  return val;
}

#define SIMPLIFIED(E) (((E)==NULL) || ((E)->type == EXPR_NUMBER))
#define NSIMPLIFIED(E) (((E)!=NULL) && ((E)->type != EXPR_NUMBER))

void link_expr_simplify(struct link_expr_t *expr)
{
  struct link_expr_t *a, *b, *c;
  int a1 = 0, a2 = 0, a3 = 0;
  int p;
  int (*op)(int a, int b, int c);

  if (!expr)
    return;

  switch (expr->type)
    {
    case EXPR_OPERATOR:
      a = expr->value.function.arg1;
      b = expr->value.function.arg2;
      c = expr->value.function.arg3;

      link_expr_simplify(a);
      link_expr_simplify(b);
      link_expr_simplify(c);

      if (NSIMPLIFIED(a) || NSIMPLIFIED(b) || NSIMPLIFIED(c))
        return;
      op = expr->value.function.calc;
      if (a)
	a1 = a->value.value;
      if (b)
	a2 = b->value.value;
      if (c)
	a3 = c->value.value;
      expr->value.value = op(a1, a2, a3);
      expr->type = EXPR_NUMBER;
      return;
    case EXPR_NUMBER:
      return;
    case EXPR_PAGE:
      if (!link_entry_defined(expr->value.entry))
        return; // 'entry' is not yet known
      p = expr->value.entry->page;
      expr->type = EXPR_NUMBER;
      expr->value.value = p;
      return;
    case EXPR_VALUE:
      if (!link_entry_defined(expr->value.entry))
        return; // 'expr' is not yet known
      a = expr->value.entry->expr;
      link_expr_simplify(a);
      *expr = *a; // copy value of 'a' to 'expr'
      return;
    default:
      ERROR("internal error\n");
    }
}

void link_expr_print_unresolved(struct link_expr_t *expr)
{
  if (!expr)
    return;
  switch (expr->type)
    {
    case EXPR_OPERATOR:
      link_expr_print_unresolved(expr->value.function.arg1);
      link_expr_print_unresolved(expr->value.function.arg2);
      link_expr_print_unresolved(expr->value.function.arg3);
      break;
    case EXPR_NUMBER:
      break;
    case EXPR_PAGE:
    case EXPR_VALUE:
      WARNING("unresolved label \"%s\"\n", expr->value.entry->entry.name);
      break;
    default:
      ERROR("internal error\n");
    }  
}

void link_expr_delete(struct link_expr_t *expr)
{}

// everything is freed in 1 time
void link_expr_free(struct link_expr_t *expr)
{}
